

/* 
Bibliotheque Caliper
Version 1.1
Auteur : Cordier Yves
Site : www.altituduino.com

License: GNU General Public License

V1.0 12/12/2020
		implementation de la classe CaliperBin (transmission binaire)
v1.01 17/12/2020
		rajout de la fonction isActive()
v1.02 27/12/2020
		rajout de la fonction detectUnitAuto(boolean mode) 
		la classe s'apelle désormais caliperBin6 mais par compatibilité il est possible de 
		l'invoquer sous le nom de caliperBin
v1.03 30/01/2021
		correction de la fonction (private) waitForeTrame()
		correction de la fonction (public) getValueLong() qui renvoi une valeur d'erreur > 100000 en cas de non detection de trame
		modification de la fonction isActive qui maintenant vérifie l'activité de la broche clock
		
v1.1  30/01/2021
		Creation d'une classe Shahe qui installe les bons paramètres pour les règles de cette marque
		Modification de la procedure de lecture qui génère un NonANomber NAN en cas d'échec 
		de lecture de la règle (typiquement débranchée).



Cette bibliotheque permet d'interfacer aiséement des calibres à coulisses numériques "chinois"
et, de réaliser pour une somme modique des stations de mesure.

This librarie intend to interface cheap "chines" numerical caliper in order to
make also cheap measure units.

Thank's to http://pcbheaven.com/expages/Digital_Caliper_Protocol
			https://sites.google.com/site/marthalprojects/home/arduino/arduino-reads-digital-caliper

which guides me to solve multiples problems

La classe CaliperBin implémentée permet
permet de lire les données issues d'un calibre à coulisse
"chinois" en mode binaire
pour 3€ resolution 0.1mm
https://www.wish.com/product/5d031bae5244fb48cb3c95b3?hide_login_modal=true&from_ad=goog_shopping&_display_country_code=FR&_force_currency_code=EUR&pid=googleadwords_int&c=%7BcampaignId%7D&ad_cid=5d031bae5244fb48cb3c95b3&ad_cc=FR&ad_lang=EN&ad_curr=EUR&ad_price=2.00&fallback_cids=5c4a02f61838ab2680e705055c78f802cc31bd2b552c78175bd3012c2eee4a3ef7c1ef9459819f860549dc0857deb1b45c1b5d9b72a7b560371275215bd30232a3530a1706b65486&campaign_id=10079127772&gclid=EAIaIQobChMI2Yj207XI7QIVlYjVCh1pLgcCEAQYASABEgKC1PD_BwE&share=web
pour 16€ resolution 0.01 mmm
https://www.amazon.fr/HXY-Digital-coulisse-batterie-suppl%C3%A9mentaire/dp/B01EI89LWC/ref=asc_df_B01EI89LWC/?tag=googshopfr-21&linkCode=df0&hvadid=186860376435&hvpos=&hvnetw=g&hvrand=4714705957934862862&hvpone=&hvptwo=&hvqmt=&hvdev=c&hvdvcmdl=&hvlocint=&hvlocphy=9055061&hvtargid=pla-348662350011&psc=1

**********************
L'interfacage physique
**********************

Ces appareils possèdent 4 broches de sorties de gauche à droite en regardant le pieds par l'arriere
-> le + (1.5v de la pile LR44)
-> le clock (horloge synchrone ou presque)
-> les datas
-> le - (0v de la pile LR44)

		haut coté écran
+---------------------+
|  +1.5  clk data 0v  |     0 <-molette de blocage
+---------------------+
		bas coté règle acier
		
Habituellement (sur les modèles essayés),
la sortie se fait en un train de 24 bits séparés en 6 quartés (4bits)
Le format est le suivant :
1 bit de start
19 bits de données
1 bit de signe

Il est à noter que les bits sont valides lorsque la broche horloge passe
de l'état haut à l'état bas. Au repos cette broche est à l'état haut.

Les tensions de sorties des broches ne sont pas compatibles avec les niveaux logiques des microcontrôleurs standards,
il convient donc de les réhausser à un niveau électrique convenable (3.3v ou 5v) à l'aide d'un dispositif
electronique adéquat.

Nous pouvons utiliser plusieurs méthodes :

Un transistor NPN et deux résistances
					+5v
					|
					R=10K
					|
					+----> signal
					|
					/
					|/
>clk-R=10K-------|   NPN Transistor ex : bC 548
					|\
					\
					v
					|
					0v

Dans cette méthode, les signaux sont inversés il convient donc dans le constructeur
de l'indiquer
ex : cal CaliperBin (5,6,true);

Une autre méthode plus osée (mais qui peut poser des problèmes de retour de masse
sur des circuits complexes) est de jouer sur les niveaux logiques des entrées

					+5v arduino
					|
					R=10K
					|                              +-----+
					+----> 1.5v calibre à coulisse +     |
					|                              |     +---> clock ---> entree Arduino
					R=10K                          |     |
					|                              |     +---> data  ---> entree Arduino
					+----> 0v calibre à coulisse   +     |
					|                              +-----+
					R=10K
					|
					0v arduino

avec ce montage un clock à 0 sera à 1.66v sur l'arduino normalement compris comme un 0
un clock à 1 sera à 3.22v sur l'arduino normalement compris comme 1

On élimine de ce fait la pile interne et on joue sur l'alimentation de l'arduino pour donner du 1.66v au lieu de 1.5v de
la pile au calibre à coulisse
Mais attention à ce que la partie métallique du calibre à coulisse ne soit pas reliée à une quelconque borne
de la pile. Cela pourait provoquer un court circuit en cas de contact avec un autre + ou - du circuit

Une autre solution plus élégante et comme diraient nos amis outre atlantique à la feuille d'érable, plus "safe"
est d'utiliser un module de conversion de niveau logique à base de TXS0108E 8-Ch Logic Level Converter Module.
Ce circuit vous coutera 2€ acheté via aliexpress ou autre et résoudra tous vos problemes
ex : https://protosupplies.com/product/txs0108e-8-ch-logic-level-converter-module/
Attention dans l'utilisation ne pas oublier de relier la broche OE (out Enable) au +1.5v du calibre à coulisse
Les entrées A étant celles à 1.5v et B celles à 5v ou 3.3v suivant votre microcontroleur


L'interfacage des règles SHAHE a posé plus de probèmes.
La solution qui donne le plus de satisfaction passe par l'utilisation d'un ampli opérationnel
à alimentation assymétrique Il est necessaire de recupérer les signaux à l'intérieur du boitier
afficheur shahe (on peut l'alimenter en externe à l'aide du 3.3v de l'arduino en éliminant les piles)
soit sur la prise mini usb soit sur la carte sur les deux petites pastilles cuivre de 0.5 mm de diametre
à proximité de la prise.

            +5v    
	          |
		   +------+
Clk   -----|+     |
           |      |---- Clk vers arduino
+0,8v -----|-     |
           +------+
              |
             0v
			
Idem pour la broche Data

Le LM741 ainsi que LM358 peuvent convenir mais leurs rapidité de reaction (0,3 V/micros) peut 
provoquer des erreurs d'interpretation du signal.
Leur préférer des modèles plus rapide avec (mini 10 V/micros)
		
************
Les méthodes
************

L'utilisateur n'a qu'a se soucier de quelques méthodes :
Le constructeur :
caliperBin(byte clockPin, byte dataPin, boolean reverseSignal = false) Le dernier arguement est optionnel faux par defaut

Le destructeur :
	~CaliperBin()

Changement de resolution (notament pour les pouces
setResolution(float r = 0.01) pour une mesure en pouce on mettra 0.005

Modification des caractéristiques de la trame (à n'utiliser que pour des calibres à coulisses encore plus exotiques que ...)
setFormat(boolean start = true, byte data = 19, byte posSigne = 21, byte posUnit = 23)

prise des valeurs , decodage et renvoi

nombre de points (>0 ou <0)
long getValueLong() renvoi la position sur un long, renvoi (2147483647l) en cas d'érreur de lecture

valeur réelle prenant en compte la résolution
float getValue() renvoi un NAN (not a number) si la règle est débranchée

savoir si la regle n'est pas eteinte 
boolean isActive()
Nota : 	cette dernière fonction est fiable à 100% si elle retourne false
			mais dans le cas de true, la règle peut tomber en "someil" juste apres le test !

**************************************************************************
Les signaux venant du calibre à coulisse tels que mesurés à l'oscilloscope
**************************************************************************

La validation du bit se fait sur le front descendant de l'horloge

: L'horloge
							<----- 10 ms ----------------------------------------------------> pour les 6 quartes
		en microseconde       <12 >                <---59---->
niveau 1  ------------------+  +--+  +--+  +--+  +---------+  +--+  +--+  +--+  +------
							|  |  |  |  |  |  |  |         |  |  |  |  |  |  |  |       6 quartés
							v  |  v  |  v  |  v  |         v  |  v  |  v  |  v  |
niveau 0                    +--+  +--+  +--+  +--+         +--+  +--+  +--+  +--+

							c     c     c     c            c     c     c     c
							0     1     2     3            4     5     6     7

: les datas exemple
niveau 1                   +--+        +--+--+--+
niveau 0 ------------------+  +--+--+--+        +---------+  +---+  -
							s     d     d     d            d      d
									0     1     2            3      4
							=     =     =     =
							1     0     1     1

*/
#ifndef _Caliper_H_File
#define _Caliper_H_File

#include <Arduino.h>

// utilisation d'une bibliotheque tiers (ou presque !)
#include <FastInOut.h> //à télécharger sur www.altituduino.com

// Les données immuables (ou presque !!)
// *************************************
#define _caliperDelaiEntreDeuxTramesMini (20000l) // delai en micosecondes entre 2 trames au mini (20 ms pour mes appareils)
// à modifier après etude à l'osciloscope pour un calibre à coulisse tres exotique

#define _mmResolution (0.01)
#define _inchResolution (0.0005)
// utilisées pour des règles dont on détetecte de manière automatique l'unité

// Quelques valeurs pour gérer les erreurs de transmission
#define _valeurErreurLecture (2147483647l) // valeur maxi d'un long
#define _CaliperNombreMaxiBouclesAttente 100000ul


// La class maitresse (la classe !! comme diraient les ados)
// ********************************************************
class CaliperBin6
{

	private :
	
	// Les données internes à l'objet
	FastInOut *clk = NULL; // entrée clock utilisant la bibliothèque FastInOut à télécharger sur www.altituduino.com
	FastInOut *data = NULL; // entrée data utilisant la bibliotheque FastInout à télécharger sur www.altituduino.com
	boolean bas = false; // correspondance entre le niveau informatique bas et le niveau electrique bas
	byte sizeMessage ; // taille du train de bit
	boolean startBit ; // existance d'un start bit
	byte  numData ; // nombre de bits de données hors start et signe
	int  signBit ; // >0 donne la position du signe ; <0 indique sans signe
	int  unitBit ; // >0 donne la position du bit indiquant l'unité (mm ou inch) habituellement 24ème bit => 23 
	long masqueSaisie; // pour isoler les numData bits de données
	float resolution =1; // resolution
	

	// Les méthodes internes à l'objet
	// *************************************************************
	// Calcul le temps entre deux instants avec gestion du roll over
	unsigned long tempsEntre(unsigned long Tact, unsigned long Tsto)
	{
		if (Tact > Tsto) return Tact - Tsto;
		return (4294967295ul - Tsto) + Tact;
	}

	// ******************************
	// visualisation binaire d'un long (utilisée pour debogger)
	void imprime(long v)
	{
		for (int i = 0; i < 32; i++)
		{
			Serial.print((v & (1l << i)) >> i);
			Serial.print(' ');
		}
		Serial.println();
	}

	// *****************************
	// attente du debut d'un message
	boolean waitForTrame()
	{
		// on attends un calme de la broche clock  durant _caliperDelaiEntreDeuxTramesMini typiquement 20 ms
		boolean posClk=clk->read();
		boolean posClkSto=posClk;
		unsigned long T = micros();
		do
		{
			posClk=clk->read();
			if (posClk && ! posClkSto)
			{ 
				T=micros(); // on enregistre la modif
				posClkSto=posClk; // mise à jour
			}
		}
		while (tempsEntre(micros(), T) < _caliperDelaiEntreDeuxTramesMini);
		
		return (posClk!=bas); // la broche est haute si le dispositif (regle et adaptateur) est presente basse si non
	}

	public :
	// Les méthodes accessibles par l'utilisateur

	// ***************
	// Le constructeur
	CaliperBin6 (byte pinClk, byte pinData, boolean reverse = false)
	{
		clk = new FastInOut(pinClk, INPUT_PULLUP); // l'entree horloge
		data = new FastInOut(pinData, INPUT_PULLUP); // l'entree data
		bas = reverse;
		setFormat(); // format standard
		setResolution(); // resolution standard
	}
	// **************
	// Le destructeur
	~CaliperBin6()
	{
		delete clk;
		delete data;
	}


	// ****************************
	// modification de la resolution
	void setResolution(float r = 0.01)
	{
		resolution = r;
	}

	// *********************************************
	// modification des caractéristiques de la trame

	void setFormat(boolean start = true, byte data = 19, byte posSigne = 21,byte posUnit = 23) // 21 : 22 eme bit 23: 24eme bit
	{
		sizeMessage = 24;
		startBit = start;
		numData = data;
		signBit = posSigne;
		unitBit=posUnit;
		masqueSaisie = 0;
		// génération du masque de saisie
		for (int i = 0; i < numData; i++) masqueSaisie |= (1l << i);

	}

	// detection si la regle est active
	// permet de déterminer à coup sur si elle n'est pas active mais
	// peut s'être etteinte juste apres le test. (ce serait malchance !)
	// *****************************************************************
	boolean isActive()
	{
		// on vérifie que la broche clock  change d'état durant 300 ms au pire
		boolean posClk=clk->read();
		boolean posClkSto=posClk;
		boolean changement=false;
		unsigned long T = micros();
		do
		{
			posClk=clk->read();
			if (posClk && ! posClkSto)
			{ 
				T=micros(); // on enregistre la modif
				changement=true;
			}
		}
		while ((tempsEntre(micros(), T) < _caliperDelaiEntreDeuxTramesMini) && !changement);
		
		return changement; 
		
	}	  
	
	// mise en detection automatique ou non d'unité
	// **********************************************
	// dans le cas non detection, il appartient à l'utilisateur de mettre la& resolution
	// via la commande setResolution(val) 
	// en mode automatique l'unité est détectée via le 24 eme bit et la resolution est automatiquement
	// mise soit à 0.01 (mm) (précision de 1/100émé de mm) soit à 0.0005 (inch) (precision du /2000 de pouce)
	void detectUnitAuto(boolean mode=true)
	{
		if (mode) unitBit=abs(unitBit);
		else unitBit=-abs(unitBit);
	}
	// prise des valeurs binaires  et decodage de celles ci
	// renvoi la valeur affichée sous forme d'un reel
	// *****************************************************

	long getValueLong()

	{
		if (waitForTrame()) // attente de la possibilité d'une trame
		{
			long v=0;
			// acquisition de la trame bit à bit
			unsigned long n=0;
			for (byte j = 0; j < sizeMessage; j++)
			{   n=0;
				while ((clk->read() != bas) && (n<_CaliperNombreMaxiBouclesAttente)){n++;} ; // On attends que le clk redescende
				if (n>=_CaliperNombreMaxiBouclesAttente) return _valeurErreurLecture; // sortie en erreur
				if (data->read() != bas) v |= (1l << j); // prise du bit de données
				n=0;
				while ((clk->read() == bas)&& (n<_CaliperNombreMaxiBouclesAttente)) {n++;} ;// On attends la remonté
				if (n>=_CaliperNombreMaxiBouclesAttente) return _valeurErreurLecture; // sortie en erreur
			}
			// traitement de la trame
			boolean signe = false; // positif par defaut
			if (signBit > 0) {signe = (v & (1l << signBit)) != 0;};
			
			if (unitBit > 0) 
			{
				if ((v & (1l<< unitBit))!=0)  resolution=_inchResolution;
				else resolution=_mmResolution;
			}
			if (startBit) v = v >> 1 ;
			v &= masqueSaisie; // recupération de la donnée numérique
			if (signe) return v = -v; // cas negatif
			return v;	
		}
		return  _valeurErreurLecture;// un long de 32 bits en erreur par defaut (valeur tres grande)
	}


	float getValue()
	{
		long v = getValueLong(); // un long de 32 bits
		if (_valeurErreurLecture == v ){return NAN;}
		return (float)v * resolution;
	}

};

// Une classe dérivée pour les règles shahe
class Shahe : public CaliperBin6 
{
	public :
	Shahe (byte pinClk, byte pinData, boolean reverse = false) : CaliperBin6(pinClk,pinData,reverse)
	{
		setFormat( true,  19,  21,0); // pas de bits d'unités (0)
		setResolution(-0.01); // negatif pour etre en coherence avec l'aficheur de la règle (bizarerie chinoise ou chinoiserie ??)
	}
};

#endif // du #define _Caliper_H_File
